
#===== 3DS Games =====#

class Friends3DS:
	TITLE_ID_EUR = 0x0004013000003202
	TITLE_ID_USA = 0x0004013000003202
	TITLE_ID_JAP = 0x0004013000003202
	LATEST_VERSION = 20

	# Friends 3DS has no product code
	PRODUCT_CODE_EUR = "----"
	PRODUCT_CODE_USA = "----"
	PRODUCT_CODE_JAP = "----"
	
	GAME_SERVER_ID = 0x3200
	ACCESS_KEY = "ridfebb9"
	NEX_VERSION = 20000


#===== Wii U Games =====#

class Friends:
	TITLE_ID_EUR = 0x10001C00
	TITLE_ID_USA = 0x10001C00
	TITLE_ID_JAP = 0x10001C00
	LATEST_VERSION = 0
	
	GAME_SERVER_ID = 0x3200
	ACCESS_KEY = "ridfebb9"
	NEX_VERSION = 20000

class DKCTF:
	TITLE_ID_EUR = 0x0005000010138300
	TITLE_ID_USA = 0x0005000010137F00
	TITLE_ID_JAP = 0x0005000010144800
	LATEST_VERSION = 17
	
	GAME_SERVER_ID = 0x10144800
	ACCESS_KEY = "7fcf384a"
	NEX_VERSION = 30400 #3.4.0
	
class Minecraft:
	TITLE_ID_EUR = 0x00050000101D7500
	TITLE_ID_USA = 0x00050000101D9D00
	TITLE_ID_JAP = 0x00050000101DBE00
	LATEST_VERSION = 688
	
	GAME_SERVER_ID = 0x101D9D00
	ACCESS_KEY = "f1b61c8e"
	NEX_VERSION = 30901 #3.9.1
	
class MK8:
	TITLE_ID_EUR = 0x000500001010ED00
	TITLE_ID_USA = 0x000500001010EC00
	TITLE_ID_JAP = 0x000500001010EB00
	LATEST_VERSION = 64
	
	GAME_SERVER_ID = 0x1010EB00
	ACCESS_KEY = "25dbf96a"
	NEX_VERSION = 30504 #3.5.4 (AMK patch)
	
class PokkenTournament:
	TITLE_ID_EUR = 0x00050000101DF400
	TITLE_ID_USA = 0x00050000101DF500
	TITLE_ID_JAP = 0x00050000101C5800
	LATEST_VERSION = 64
	
	GAME_SERVER_ID = 0x101C5800
	ACCESS_KEY = "6ef3adf1"
	NEX_VERSION = 31000
	
class Puddle:
	TITLE_ID_EUR = 0x000500001010FB00
	TITLE_ID_USA = 0x0005000010110500
	LATEST_VERSION = 16
	
	GAME_SERVER_ID = 0x1010FB00
	ACCESS_KEY = "afcffb5c"
	NEX_VERSION = 30001 #3.0.1.4

class SMM:
	TITLE_ID_EUR = 0x000500001018DD00
	TITLE_ID_USA = 0x000500001018DC00
	TITLE_ID_JAP = 0x000500001018DB00
	LATEST_VERSION = 272
	
	GAME_SERVER_ID = 0x1018DB00
	ACCESS_KEY = "9f2b4678"
	NEX_VERSION = 30810 #3.8.10 (AMA patch)
	

#===== Switch Games =====#

class ACNH:
	TITLE_ID = 0x01006F8002326000
	LATEST_VERSION = 0x1A0000
	
	GAME_SERVER_ID = 0x2EE2E300
	ACCESS_KEY = "v43a10em"
	NEX_VERSION = 40604 #4.6.4
	CLIENT_VERSION = 2
	
class GBG:
	TITLE_ID = 0x0100FA5010788000
	LATEST_VERSION = 0x10000
	
	GAME_SERVER_ID = 0x2E99DD00
	ACCESS_KEY = "97b08aad"
	NEX_VERSION = 40610
	CLIENT_VERSION = 2

class MK8Deluxe:
	TITLE_ID = 0x0100152000022000
	LATEST_VERSION = 0xA0000
	
	GAME_SERVER_ID = 0x2B309E01
	ACCESS_KEY = "09c1c475"
	NEX_VERSION = 40302

class SMM2:
	TITLE_ID = 0x01009B90006DC000
	LATEST_VERSION = 0x50000
	
	GAME_SERVER_ID = 0x22306D00
	ACCESS_KEY = "fdf6617f"
	NEX_VERSION = 40605 #4.6.25 (appslop)
	CLIENT_VERSION = 60
	
	PIA_VERSION = 51898
	PIA_APP_VERSION = 3
	PIA_KEY = bytes.fromhex("667c18475889faab61f93ef1da180971")

class Splatoon2:
	TITLE_ID = 0x0100F8F0000A2000
	LATEST_VERSION = 0x310000
	
	GAME_SERVER_ID = 0x2DF33D01
	ACCESS_KEY = "4eb18d39"
	NEX_VERSION = 40311
	
	PIA_VERSION = 50903
	PIA_KEY = bytes.fromhex("ee182a63e216cdb1f51ad4bed8cf6508")

class Tetris99:
	TITLE_ID = 0x010040600C5CE000
	LATEST_VERSION = 0x60000
	
	GAME_SERVER_ID = 0x23BDA200
	ACCESS_KEY = "cdd6114d"
	NEX_VERSION = 40600 #4.6.4 (app99)
	CLIENT_VERSION = 5





WIIU = 0
SWITCH = 1


class Settings:
	MODE_ECB = 0
	MODE_GCM = 1
	
	field_types = {
		"pia.version": int,
		"pia.application_version": int,
		"pia.lan_version": int,
		
		"crypto.session_key": bytes,
		
		"common.pid_size": int
	}
	
	def __init__(self, version, app_version=-1, *, platform=SWITCH):
		self.settings = {}
		
		version //= 100
		self["pia.version"] = version
		self["pia.application_version"] = app_version
		self["pia.lan_version"] = self.lan_version(version)
		
		self["common.pid_size"] = 8 if platform == SWITCH else 4
	
	def __getitem__(self, name): return self.settings[name]
	def __setitem__(self, name, value):
		if name not in self.field_types:
			raise KeyError("Unknown setting: %s" %name)
		self.settings[name] = self.field_types[name](value)
	
	def lan_version(self, version):
		if version < 509: return 0 # No crypto challenge
		if version < 511: return 1 # Crypto challenge
		return 2 # Crypto challenge and IPv6 support

def default(version, app_version=-1, *, platform=SWITCH):
	return Settings(version, app_version, platform=platform)




from Crypto.Util.Padding import pad
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Hash import SHA256
from Crypto.Random import get_random_bytes
from anynet import tls, http
from nintendo import switch
import pkg_resources
import struct

import logging
logger = logging.getLogger(__name__)


CA = pkg_resources.resource_filename("nintendo", "files/cert/CACERT_NINTENDO_CA_G3.der")


RSA_MODULUS =  int(
	"2903599220185509629948246004681271806662185201109683699434876284"
	"9306378942456577312580648895443616535088601867223713942187399041"
	"4854872772034425863387471719375473684104853507768450203982749294"
	"8967945831738920699512732919046223059402955082098739033920491649"
	"6879108565068863591362496844602988110766097564477545097467537357"
	"3183749964356608012071405755940871989007021731074728723470759285"
	"8155278592486462922448753256166071381157786436864032235809243318"
	"2591089355065715995209202752330511548478896205428626608544326862"
	"4782505679727111312756904637828438785474471375120478991907979820"
	"98870089661523253199182993983393803812441"
)

RSA_EXPONENT = 65537


USER_AGENT = {
	 900: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 901: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 910: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 920: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	1000: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1001: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1002: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1003: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1004: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1010: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1011: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1020: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1100: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1101: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1200: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1201: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1202: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1203: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1210: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1300: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.3.0.0; Add-on 13.3.0.0)",
	1310: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1320: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1321: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1400: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1410: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1411: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1412: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
}

LATEST_VERSION = 1412


class AAuthError(switch.NDASError): pass


class AAuthClient:
	def __init__(self):
		self.url = "aauth-lp1.ndas.srv.nintendo.net"
		
		self.user_agent = USER_AGENT[LATEST_VERSION]
		self.power_state = "FA"
		
		ca = tls.TLSCertificate.load(CA, tls.TYPE_DER)
		self.context = tls.TLSContext()
		self.context.set_authority(ca)
	
	def set_url(self, url): self.url = url
	def set_user_agent(self, user_agent): self.user_agent = user_agent
	def set_power_state(self, state): self.power_state = state
	
	def set_context(self, context):
		self.context = context
	
	def set_system_version(self, version):
		if version not in USER_AGENT:
			raise ValueError("Unknown system version")
		self.user_agent = USER_AGENT[version]
	
	async def request(self, req, use_power_state):
		req.headers["Host"] = self.url
		req.headers["User-Agent"] = self.user_agent
		req.headers["Accept"] = "*/*"
		if use_power_state:
			req.headers["X-Nintendo-PowerState"] = self.power_state
		req.headers["Content-Length"] = 0
		req.headers["Content-Type"] = "application/x-www-form-urlencoded"
		
		response = await http.request(self.url, req, self.context)
		if response.error():
			if response.json:
				logger.error("AAuth server returned errors:")
				errors = response.json["errors"]
				for error in errors:
					logger.error("  (%s) %s", error["code"], error["message"])
				raise AAuthError(response.status_code, errors)
			else:
				logger.error("AAuth server returned status code %i", response.status_code)
				raise AAuthError(response.status_code)
		return response
		
	def verify_ticket(self, ticket, title_id):
		if len(ticket) != 0x2C0:
			raise ValueError("Ticket has unexpected size")
		if struct.unpack_from("<I", ticket)[0] != 0x10004:
			raise ValueError("Ticket has invalid signature type")
		if struct.unpack_from(">Q", ticket, 0x2A0)[0] != title_id:
			raise ValueError("Ticket has different title id")
		if struct.unpack_from(">Q", ticket, 0x2A8)[0] != ticket[0x285]:
			raise ValueError("Ticket has inconsistent master key revision")

	async def get_time(self):
		req = http.HTTPRequest.get("/v1/time")
		req.headers["Host"] = self.url
		req.headers["Accept"] = "*/*"
		
		response = await http.request(self.url, req, self.context)
		response.raise_if_error()
		
		time = int(response.headers["X-NINTENDO-UNIXTIME"])
		ip = response.headers["X-NINTENDO-GLOBAL-IP"]
		return time, ip
	
	async def challenge(self, device_token):
		req = http.HTTPRequest.post("/v3/challenge")
		req.rawform = {
			"&device_auth_token": device_token
		}
		
		response = await self.request(req, False)
		return response.json

	# Warning: do not use auth_nocert on a production server.
	# It will immediately ban your Switch.
	async def auth_nocert(self, title_id, title_version, device_token):
		req = http.HTTPRequest.post("/v3/application_auth_token")
		req.form = {
			"application_id": "%016x" %title_id,
			"application_version": "%08x" %title_version,
			"device_auth_token": device_token,
			"media_type": "NO_CERT"
		}
		
		response = await self.request(req, True)
		return response.json

	async def auth_system(self, title_id, title_version, device_token):
		req = http.HTTPRequest.post("/v3/application_auth_token")
		req.form = {
			"application_id": "%016x" %title_id,
			"application_version": "%08x" %title_version,
			"device_auth_token": device_token,
			"media_type": "SYSTEM"
		}

		response = await self.request(req, True)
		return response.json

	async def auth_digital(self, title_id, title_version, device_token, ticket):
		self.verify_ticket(ticket, title_id)
		
		plain_key = get_random_bytes(16)
		
		aes = AES.new(plain_key, AES.MODE_CBC, iv=bytes(16))
		encrypted_ticket = aes.encrypt(pad(ticket, 16))
		
		rsa_key = RSA.construct((RSA_MODULUS, RSA_EXPONENT))
		rsa = PKCS1_OAEP.new(rsa_key, SHA256)
		encrypted_key = rsa.encrypt(plain_key)
	
		req = http.HTTPRequest.post("/v3/application_auth_token")
		req.form = {
			"application_id": "%016x" %title_id,
			"application_version": "%08x" %title_version,
			"device_auth_token": device_token,
			"media_type": "DIGITAL",
			
			"cert": switch.b64encode(encrypted_ticket),
			"cert_key": switch.b64encode(encrypted_key)
		}
		
		response = await self.request(req, True)
		return response.json




from nintendo.pia import packet, streams
import copy


FLAG_MULTICAST = 1


class Message:
	def __init__(self):
		self.flags = 0
		self.station_index = 0xFD
		self.destination = 0
		self.source = 0
		self.protocol_id = 0
		self.protocol_port = 0
		self.payload = b""


class MessageEncoderV1:
	def __init__(self, settings):
		self.settings = settings
	
	def decode(self, data):
		stream = streams.StreamIn(data, self.settings)
		
		messages = []
		while not stream.eof():
			message = Message()
			message.flags = stream.u8()
			message.station_index = stream.u8()
			size = stream.u16()
			message.destination = stream.u32()
			message.source = stream.u32()
			message.protocol_id = stream.u16()
			message.protocol_port = stream.u16()
			stream.pad(4)
			message.payload = stream.read(size)
			stream.align(4)
			messages.append(message)
		return messages
	
	def encode(self, messages):
		stream = streams.StreamOut(self.settings)
		for message in messages:
			stream.u8(message.flags)
			stream.u8(message.station_index)
			stream.u16(len(message.payload))
			stream.u32(message.destination)
			stream.u32(message.source)
			stream.u16(message.protocol_id)
			stream.u16(message.protocol_port)
			stream.pad(4)
			stream.write(message.payload)
			stream.align(4)
		return stream.get()


class MessageEncoderV2:
	def __init__(self, settings):
		self.settings = settings
	
	def decode(self, data):
		stream = streams.StreamIn(data, self.settings)
		
		messages = []
		while not stream.eof():
			message = Message()
			message.flags = stream.u8()
			size = stream.u16()
			message.destination = stream.u64()
			message.source = stream.u64()
			message.protocol_id = stream.u8()
			message.protocol_port = stream.u8()
			stream.pad(3)
			message.payload = stream.read(size)
			stream.align(4)
			messages.append(message)
		return messages
	
	def encode(self, messages):
		stream = streams.StreamOut(self.settings)
		for message in messages:
			stream.u8(message.flags)
			stream.u16(len(message.payload))
			stream.u64(message.destination)
			stream.u64(message.source)
			stream.u8(message.protocol_id)
			stream.u8(message.protocol_port)
			stream.pad(3)
			stream.write(message.payload)
			stream.align(4)
		return stream.get()


class MessageEncoderV3:
	def __init__(self, settings):
		self.settings = settings
	
	def decode(self, data):
		stream = streams.StreamIn(data, self.settings)
		
		messages = []
		while not stream.eof():
			message = Message()
			message.flags = stream.u8()
			if stream.u8() != 1:
				raise ValueError("Message has wrong version number")
			size = stream.u16()
			message.protocol_id = stream.u8()
			message.protocol_port = stream.u8()
			message.destination = stream.u64()
			message.source = stream.u64()
			message.payload = stream.read(size)
			stream.align(4)
			messages.append(message)
		return messages
	
	def encode(self, messages):
		stream = streams.StreamOut(self.settings)
		for message in messages:
			stream.u8(message.flags)
			stream.u8(1)
			stream.u16(len(message.payload))
			stream.u8(message.protocol_id)
			stream.u8(message.protocol_port)
			stream.u64(message.destination)
			stream.u64(message.source)
			stream.write(message.payload)
			stream.align(4)
		return stream.get()


class MessageEncoderV4:
	def __init__(self, settings):
		self.settings = settings
	
	def decode(self, data):
		stream = streams.StreamIn(data, self.settings)
		
		messages = []
		while not stream.eof():
			message = Message()
			message.flags = stream.u8()
			if stream.u8() != 2:
				raise ValueError("Message has wrong version number")
			size = stream.u16()
			message.protocol_id = stream.u8()
			message.protocol_port = stream.u24()
			message.destination = stream.u64()
			message.source = stream.u64()
			message.payload = stream.read(size)
			stream.align(4)
			messages.append(message)
		return messages
	
	def encode(self, messages):
		stream = streams.StreamOut(self.settings)
		for message in messages:
			stream.u8(message.flags)
			stream.u8(2)
			stream.u16(len(message.payload))
			stream.u8(message.protocol_id)
			stream.u24(message.protocol_port)
			stream.u64(message.destination)
			stream.u64(message.source)
			stream.write(message.payload)
			stream.align(4)
		return stream.get()


class MessageEncoderV5:
	def __init__(self, settings):
		self.settings = settings
	
	def decode(self, data):
		stream = streams.StreamIn(data, self.settings)
		
		messages = []
		
		size = None
		message = Message()
		while not stream.eof():
			flags = stream.u8()
			
			if flags & 1: message.flags = stream.u8()
			if flags & 2: size = stream.u16()
			if flags & 4:
				message.protocol_id = stream.u8()
				message.protocol_port = stream.u24()
			if flags & 8: message.destination = stream.u64()
			if flags & 16: message.source = stream.u64()
			
			message.payload = stream.read(size)
			stream.align(4)
			
			messages.append(message)
			message = copy.copy(message)
		
		return messages
	
	def encode(self, messages):
		stream = streams.StreamOut(self.settings)
		
		prev = Message()
		for message in messages:
			flags = 0
			if message.flags != prev.flags: flags |= 1
			if len(message.payload) != len(prev.payload): flags |= 2
			if message.protocol_id != prev.protocol_id or message.protocol_port != prev.protocol_port:
				flags |= 4
			if message.destination != prev.destination: flags |= 8
			if message.source != prev.source: flags |= 16
			
			stream.u8(flags)
			if flags & 1: stream.u8(message.flags)
			if flags & 2: stream.u16(len(message.payload))
			if flags & 4:
				stream.u8(message.protocol_id)
				stream.u24(message.protocol_port)
			if flags & 8: stream.u64(message.destination)
			if flags & 16: stream.u64(message.source)
			
			stream.write(message.payload)
			stream.align(4)
			
			prev = message
		return stream.get()


class MessageEncoderV6:
	def __init__(self, settings):
		self.settings = settings
	
	def decode(self, data):
		stream = streams.StreamIn(data, self.settings)
		
		messages = []
		
		size = None
		message = Message()
		while not stream.eof():
			flags = stream.u8()
			
			if flags & 1: message.flags = stream.u8()
			if flags & 2: size = stream.u16()
			if flags & 4:
				message.protocol_id = stream.u8()
				message.protocol_port = stream.u24()
			if flags & 8: message.destination = stream.u64()
			
			message.payload = stream.read(size)
			stream.align(4)
			
			messages.append(message)
			message = copy.copy(message)
		
		return messages
	
	def encode(self, messages):
		stream = streams.StreamOut(self.settings)
		
		prev = Message()
		for message in messages:
			flags = 0
			if message.flags != prev.flags: flags |= 1
			if len(message.payload) != len(prev.payload): flags |= 2
			if message.protocol_id != prev.protocol_id or message.protocol_port != prev.protocol_port:
				flags |= 4
			if message.destination != prev.destination: flags |= 8
			
			stream.u8(flags)
			if flags & 1: stream.u8(message.flags)
			if flags & 2: stream.u16(len(message.payload))
			if flags & 4:
				stream.u8(message.protocol_id)
				stream.u24(message.protocol_port)
			if flags & 8: stream.u64(message.destination)
			
			stream.write(message.payload)
			stream.align(4)
			
			prev = message
		return stream.get()


class MessageEncoder:
	def __init__(self, settings):
		self.encoder = self.make_encoder(settings)
	
	def make_encoder(self, settings):
		version = settings["pia.version"]
		if version <= 504: return MessageEncoderV1(settings)
		elif 506 <= version <= 510: return MessageEncoderV2(settings)
		elif 511 <= version <= 512: return MessageEncoderV3(settings)
		elif 514 <= version <= 517: return MessageEncoderV4(settings)
		elif 518 <= version <= 526: return MessageEncoderV5(settings)
		elif 527 <= version <= 616: return MessageEncoderV6(settings)
		else:
			raise ValueError("Unsupported pia version")
	
	def decode(self, data): return self.encoder.decode(data)
	def encode(self, messages): return self.encoder.encode(messages)


class MessageTransport:
	def __init__(self, transport, settings):
		self.transport = transport
		self.encoder = MessageEncoder(settings)
	
	async def send(self, messages, address, encrypt=True):
		pack = packet.Packet()
		pack.payload = self.encoder.encode(messages)
		pack.encrypted = encrypt
		await self.transport.send(pack, address)
	
	async def receive(self):
		packet, address = await self.transport.receive()
		try:
			messages = self.encoder.decode(packet.payload)
			return messages, address
		except Exception:
			logger.warning("Received invalid message buffer")


class MessageHandler:
	def __init__(self, transport, stations, settings):
		self.transport = MessageTransport(transport, settings)
		self.stations = stations
	
	async def send(self, message, address, encrypt=True):
		local = self.stations.local_station()
		message.station_index = local.index
		message.source = local.constant_id
		await self.transport.send([message], address, encrypt)
	
	async def send_station(self, message, index, encrypt=True):
		station = self.stations.find_by_index(index)
		message.flags &= ~FLAG_MULTICAST
		message.destination = station.constant_id
		await self.send(message, station.address, encrypt)




from Crypto.Hash import CMAC
from Crypto.Cipher import AES
from anynet import tls, http
from nintendo import switch
import pkg_resources

import logging
logger = logging.getLogger(__name__)


CA = pkg_resources.resource_filename("nintendo", "files/cert/CACERT_NINTENDO_CA_G3.der")


DAUTH_SOURCE = bytes.fromhex("8be45abcf987021523ca4f5e2300dbf0")

SYSTEM_VERSION_DIGEST = {
	900:  "CusHY#00090000#-80vwBkUjWLb5Kpb_cnuTjBZ0rHwZHhN7R1-vg0Ti5c=",
	901:  "CusHY#00090001#qVDSOCehwMDCHyDnkXiTSJ1wEJZHtpRV_CLMKgD-fSw=",
	910:  "CusHY#00090100#vIPNrRbf30SoU8ZJ6uGklMqKAkyjHfdE9m6yLFeChkE=",
	920:  "CusHY#00090200#Uxxmc8gYnfMqxzdZdygZ_OrKo98O7QA65s_EkZnGsDo=",
	1000: "CusHY#000a0000#EmdxOnZjZ9Ihf3Zskt_48pYgowAUeeJccU6tCBIweEc=",
	1001: "CusHY#000a0001#JEuSEdid24qqHqQzfW1tuNsCGcCk-86gcPq0I7M1x18=",
	1002: "CusHY#000a0002#BTOGo0giC7bbrNoi8JEm-FBzmXb2Kgpq4K3OzQrD5l8=",
	1003: "CusHY#000a0003#4mBbTFYnE0Rtmh8NLCVq61rbvx0kJPQUxXkDpwj0V84=",
	1004: "CusHY#000a0003#4mBbTFYnE0Rtmh8NLCVq61rbvx0kJPQUxXkDpwj0V84=",
	1010: "CusHY#000a0100#Vlw9dIEqjxE2F5jDOQPYWXs2p7wIGyDYWXXIyQGdxcE=",
	1011: "CusHY#000a0100#Vlw9dIEqjxE2F5jDOQPYWXs2p7wIGyDYWXXIyQGdxcE=",
	1020: "CusHY#000a0200#90k0dE_eO7hRcs6ByTZMvgUm4lhEoqAlik96WkznQcQ=",
	1100: "CusHY#000b0000#VyA0fsWi6ZBEOzVsseXIcEfFLqQMgW0tWzN2oJ7viqk=",
	1101: "CusHY#000b0001#iI0rZ0Q2dg3Evhd-8GoYmp-KTE8malKe0GOJgXa-XG8=",
	1200: "CusHY#000c0000#C-BynYNPXdQJNBZjx02Hizi8lRUSIKLwPGa5p8EY1uo=",
	1201: "CusHY#000c0001#YXsU5FTbkUh18QH27L3woGqw5n1gIDpMGbPXM8oACzY=",
	1202: "CusHY#000c0002#6tB3UVnmvT_nsNBMPSD-K1oe0IA1cYvYDyqDCjy2W_I=",
	1203: "CusHY#000c0003#E8Ph6vISWsJtN0E3hsfVRoZMG1Qqkc-qGRlAp-Bs2SI=",
	1210: "CusHY#000c0100#Hzs8Gtp6yKgGKMb732-5Q-NvbQcHCgBh_LQrrpg0bIs=",
	1300: "CusHY#000d0000#r1xneESd4PiTRYIhVIl0bK1ST5L5BUmv_uGPLqc4PPo=",
	1310: "CusHY#000d0100#plps6S3C43QHhkI2oNvYIFjNxQjTcLdUX2_biEI5w2w=",
	1320: "CusHY#000d0200#JFVNVuG9x3V5tRshdD9FdJjgHOmzsrgXHocEPvW5eMM=",
	1321: "CusHY#000d0201#V1i7M7oEhkDaH1lcGlHhGAnyHONMAnTAA6pGdZ7MFqc=",
	1400: "CusHY#000e0000#35hWb15SBXTnbUfTMLBz9sCnfheuRGis0OTZqa7l8yw=",
	1410: "CusHY#000e0100#ctIxSPR4jenzQNGc6y4zXIvzvF75ty53jN0T15Rjtmk=",
	1411: "CusHY#000e0101#uTt4IVydkYqwYArOFR3BzOCmw0MkEeF_tZxHENYDh4E=",
	1412: "CusHY#000e0102#jHk6_VwXVPPl3ijRZ5jRy5MIAcUW_Q2uFdfJ0vrjhCA=",
}

USER_AGENT = {
	900:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	901:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	910:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	920:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	1000: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1001: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1002: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1003: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1004: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1010: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1011: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1020: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1100: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 11.4.0.0)",
	1101: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 11.4.0.0)",
	1200: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1201: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1202: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1203: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1210: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1300: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.3.0.0)",
	1310: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.4.0.0)",
	1320: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.4.0.0)",
	1321: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.4.0.0)",
	1400: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
	1410: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
	1411: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
	1412: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
}

KEY_GENERATION = {
	900:  10,
	901:  10,
	910:  11,
	920:  11,
	1000: 11,
	1001: 11,
	1002: 11,
	1003: 11,
	1004: 11,
	1010: 11,
	1011: 11,
	1020: 11,
	1100: 11,
	1101: 11,
	1200: 11,
	1201: 11,
	1202: 11,
	1203: 11,
	1210: 11,
	1300: 13,
	1310: 13,
	1320: 13,
	1321: 13,
	1400: 14,
	1410: 14,
	1411: 14,
	1412: 14,
}

API_VERSION = {
	900: 6,
	901: 6,
	910: 6,
	920: 6,
	1000: 6,
	1001: 6,
	1002: 6,
	1003: 6,
	1004: 6,
	1010: 6,
	1011: 6,
	1020: 6,
	1100: 6,
	1101: 6,
	1200: 6,
	1201: 6,
	1202: 6,
	1203: 6,
	1210: 6,
	1300: 7,
	1310: 7,
	1320: 7,
	1321: 7,
	1400: 7,
	1410: 7,
	1411: 7,
	1412: 7,
}

LATEST_VERSION = 1412


class DAuthError(switch.NDASError): pass


class DAuthClient:
	SCSI = 0x146C8AC7B8A0DB52
	ATUM = 0x3117B250CAB38F45
	ESHOP = 0x41F4A6491028E3C4
	BCAT = 0x67BF9945B45248C6
	SATA = 0x6AC5A6873FE5F68C
	ACCOUNT = 0x81333C548B2E876D
	NPNS = 0x83B72B05DC3278D7
	BAAS = 0x8F849B5D34778D8E
	BEACH = 0x93AF0ACB26258DE9
	DRAGONS = 0xD5B6CAC2C1514C56
	PCTL = 0xDC656EA03B63CF68
	PREPO = 0xDF51C436BC01C437
	
	def __init__(self, keyset):
		self.keyset = keyset
		
		ca = tls.TLSCertificate.load(CA, tls.TYPE_DER)
		self.context = tls.TLSContext()
		self.context.set_authority(ca)
		
		self.region = 1
		
		self.url = "dauth-lp1.ndas.srv.nintendo.net"
		self.user_agent = USER_AGENT[LATEST_VERSION]
		self.system_digest = SYSTEM_VERSION_DIGEST[LATEST_VERSION]
		self.key_generation = KEY_GENERATION[LATEST_VERSION]
		self.api_version = API_VERSION[LATEST_VERSION]
		
		self.power_state = "FA"
		
	def set_certificate(self, cert, key):
		self.context.set_certificate(cert, key)
	def set_context(self, context):
		self.context = context

	def set_url(self, url): self.url = url
	def set_power_state(self, state): self.power_state = state
	def set_platform_region(self, region): self.region = region
	
	def set_user_agent(self, user_agent): self.user_agent = user_agent
	def set_system_digest(self, digest): self.system_digest = digest
	def set_key_generation(self, keygen): self.key_generation = keygen
	def set_api_version(self, version): self.api_version = version
	
	def set_system_version(self, version):
		if version not in USER_AGENT:
			raise ValueError("Unknown system version: %i" %version)
		self.user_agent = USER_AGENT[version]
		self.system_digest = SYSTEM_VERSION_DIGEST[version]
		self.key_generation = KEY_GENERATION[version]
		self.api_version = API_VERSION[version]
		
	async def request(self, req):
		req.headers["Host"] = self.url
		req.headers["User-Agent"] = self.user_agent
		req.headers["Accept"] = "*/*"
		req.headers["X-Nintendo-PowerState"] = self.power_state
		req.headers["Content-Length"] = 0
		req.headers["Content-Type"] = "application/x-www-form-urlencoded"
		
		response = await http.request(self.url, req, self.context)
		if response.error():
			if response.json:
				logger.error("DAuth request returned errors:")
				errors = response.json["errors"]
				for error in errors:
					logger.error("  (%s) %s", error["code"], error["message"])
				raise DAuthError(response.status_code, errors)
			else:
				logger.error("DAuth request returned status code %i", response.status_code)
				raise DAuthError(response.status_code)
		return response
		
	async def challenge(self):
		req = http.HTTPRequest.post("/v%i/challenge" %self.api_version)
		req.form = {
			"key_generation": self.key_generation
		}
		
		response = await self.request(req)
		return response.json
	
	async def device_token(self, client_id):
		challenge = await self.challenge()
		
		data = switch.b64decode(challenge["data"])
		
		req = http.HTTPRequest.post("/v%i/device_auth_token" %self.api_version)
		req.rawform = {
			"challenge": challenge["challenge"],
			"client_id": "%016x" %client_id,
			"ist": "true" if self.region == 2 else "false",
			"key_generation": self.key_generation,
			"system_version": self.system_digest
		}
		
		string = http.formencode(req.rawform, False)
		req.rawform["mac"] = self.calculate_mac(string, data)
		
		response = await self.request(req)
		return response.json
		
	async def edge_token(self, client_id, vendor_id="akamai"):
		challenge = await self.challenge()
		
		data = switch.b64decode(challenge["data"])
		
		req = http.HTTPRequest.post("/v%i/edge_token" %self.api_version)
		req.rawform = {
			"challenge": challenge["challenge"],
			"client_id": "%016x" %client_id,
			"ist": "true" if self.region == 2 else "false",
			"key_generation": self.key_generation,
			"system_version": self.system_digest
		}
		
		if self.api_version == 7:
			req.rawform["vendor_id"] = vendor_id
		
		string = http.formencode(req.rawform, False)
		req.rawform["mac"] = self.calculate_mac(string, data)
		
		response = await self.request(req)
		return response.json
		
	def get_master_key(self):
		keygen = self.key_generation
		keyname = "master_key_%02x" %(keygen - 1)
		return self.keyset[keyname]
		
	def decrypt_key(self, key, kek):
		aes = AES.new(kek, AES.MODE_ECB)
		return aes.decrypt(key)
		
	def calculate_mac(self, form, data):
		kek_source = self.keyset["aes_kek_generation_source"]
		master_key = self.get_master_key()
		
		key = self.decrypt_key(kek_source, master_key)
		key = self.decrypt_key(DAUTH_SOURCE, key)
		key = self.decrypt_key(data, key)
		
		mac = CMAC.new(key, ciphermod=AES)
		mac.update(form.encode())
		return switch.b64encode(mac.digest())




from nintendo.sead import Random
import struct

def create_key_value(rand, table):
	value = 0
	for i in range(4):
		index = rand.uint(len(table))
		shift = rand.uint(4) * 8
		byte = (table[index] >> shift) & 0xFF
		value = (value << 8) | byte
	return value

def create_key(rand, table, size):
	if size % 4:
		raise ValueError("Key size must be multiple of 4")
	
	key = b""
	for i in range(size // 4):
		value = create_key_value(rand, table)
		key += struct.pack("<I", value)
	return key




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class MyProfile(common.Structure):
	def __init__(self):
		super().__init__()
		self.region = None
		self.country = None
		self.area = None
		self.language = None
		self.platform = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
	
	def check_required(self, settings, version):
		for field in ['region', 'country', 'area', 'language', 'platform', 'unk1', 'unk2', 'unk3']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.region = stream.u8()
		self.country = stream.u8()
		self.area = stream.u8()
		self.language = stream.u8()
		self.platform = stream.u8()
		self.unk1 = stream.u64()
		self.unk2 = stream.string()
		self.unk3 = stream.string()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u8(self.region)
		stream.u8(self.country)
		stream.u8(self.area)
		stream.u8(self.language)
		stream.u8(self.platform)
		stream.u64(self.unk1)
		stream.string(self.unk2)
		stream.string(self.unk3)


class Mii(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.mii_data = None
	
	def check_required(self, settings, version):
		for field in ['unk1', 'unk2', 'unk3', 'mii_data']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk1 = stream.string()
		self.unk2 = stream.bool()
		self.unk3 = stream.u8()
		self.mii_data = stream.buffer()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.string(self.unk1)
		stream.bool(self.unk2)
		stream.u8(self.unk3)
		stream.buffer(self.mii_data)


class MiiList(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.mii_datas = None
	
	def check_required(self, settings, version):
		for field in ['unk1', 'unk2', 'unk3', 'mii_datas']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk1 = stream.string()
		self.unk2 = stream.bool()
		self.unk3 = stream.u8()
		self.mii_datas = stream.list(stream.buffer)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.string(self.unk1)
		stream.bool(self.unk2)
		stream.u8(self.unk3)
		stream.list(self.mii_datas, stream.buffer)


class PlayedGame(common.Structure):
	def __init__(self):
		super().__init__()
		self.game_key = GameKey()
		self.datetime = None
	
	def check_required(self, settings, version):
		for field in ['datetime']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.game_key = stream.extract(GameKey)
		self.datetime = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.add(self.game_key)
		stream.datetime(self.datetime)


class GameKey(common.Data):
	def __init__(self):
		super().__init__()
		self.title_id = 0
		self.title_version = 0
	
	def check_required(self, settings, version):
		pass
	
	def load(self, stream, version):
		self.title_id = stream.u64()
		self.title_version = stream.u16()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u64(self.title_id)
		stream.u16(self.title_version)
common.DataHolder.register(GameKey, "GameKey")


class FriendMiiRequest(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk1 = None
		self.unk2 = None
	
	def check_required(self, settings, version):
		for field in ['unk1', 'unk2']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk1 = stream.u32()
		self.unk2 = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.unk1)
		stream.datetime(self.unk2)


class FriendMii(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk1 = None
		self.mii = Mii()
		self.unk2 = None
	
	def check_required(self, settings, version):
		for field in ['unk1', 'unk2']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk1 = stream.u32()
		self.mii = stream.extract(Mii)
		self.unk2 = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.unk1)
		stream.add(self.mii)
		stream.datetime(self.unk2)


class FriendMiiList(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk1 = None
		self.mii = MiiList()
		self.unk2 = None
	
	def check_required(self, settings, version):
		for field in ['unk1', 'unk2']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk1 = stream.u32()
		self.mii = stream.extract(MiiList)
		self.unk2 = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.unk1)
		stream.add(self.mii)
		stream.datetime(self.unk2)


class FriendRelationship(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
	
	def check_required(self, settings, version):
		for field in ['unk1', 'unk2', 'unk3']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk1 = stream.u32()
		self.unk2 = stream.u64()
		self.unk3 = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.unk1)
		stream.u64(self.unk2)
		stream.u8(self.unk3)


class NintendoPresence(common.Structure):
	def __init__(self):
		super().__init__()
		self.changed_bit_flag = None
		self.game_key = GameKey()
		self.game_mode_description = None
		self.join_availability_flag = None
		self.matchmake_system_type = None
		self.join_game_id = None
		self.join_game_mode = None
		self.owner_pid = None
		self.join_group_id = None
		self.application_data = None
	
	def check_required(self, settings, version):
		for field in ['changed_bit_flag', 'game_mode_description', 'join_availability_flag', 'matchmake_system_type', 'join_game_id', 'join_game_mode', 'owner_pid', 'join_group_id', 'application_data']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.changed_bit_flag = stream.u32()
		self.game_key = stream.extract(GameKey)
		self.game_mode_description = stream.string()
		self.join_availability_flag = stream.u32()
		self.matchmake_system_type = stream.u8()
		self.join_game_id = stream.u32()
		self.join_game_mode = stream.u32()
		self.owner_pid = stream.pid()
		self.join_group_id = stream.u32()
		self.application_data = stream.buffer()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.changed_bit_flag)
		stream.add(self.game_key)
		stream.string(self.game_mode_description)
		stream.u32(self.join_availability_flag)
		stream.u8(self.matchmake_system_type)
		stream.u32(self.join_game_id)
		stream.u32(self.join_game_mode)
		stream.pid(self.owner_pid)
		stream.u32(self.join_group_id)
		stream.buffer(self.application_data)


class FriendPresence(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk = None
		self.presence = NintendoPresence()
	
	def check_required(self, settings, version):
		for field in ['unk']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk = stream.u32()
		self.presence = stream.extract(NintendoPresence)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.unk)
		stream.add(self.presence)


class FriendPicture(common.Structure):
	def __init__(self):
		super().__init__()
		self.unk = None
		self.data = None
		self.datetime = None
	
	def check_required(self, settings, version):
		for field in ['unk', 'data', 'datetime']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unk = stream.u32()
		self.data = stream.buffer()
		self.datetime = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.unk)
		stream.buffer(self.data)
		stream.datetime(self.datetime)


class FriendPersistentInfo(common.Structure):
	def __init__(self):
		super().__init__()
		self.pid = None
		self.region = None
		self.country = None
		self.area = None
		self.language = None
		self.platform = None
		self.game_key = GameKey()
		self.message = None
		self.message_updated = None
		self.friended = None
		self.unk = None
	
	def check_required(self, settings, version):
		for field in ['pid', 'region', 'country', 'area', 'language', 'platform', 'message', 'message_updated', 'friended', 'unk']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.pid = stream.pid()
		self.region = stream.u8()
		self.country = stream.u8()
		self.area = stream.u8()
		self.language = stream.u8()
		self.platform = stream.u8()
		self.game_key = stream.extract(GameKey)
		self.message = stream.string()
		self.message_updated = stream.datetime()
		self.friended = stream.datetime()
		self.unk = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.pid(self.pid)
		stream.u8(self.region)
		stream.u8(self.country)
		stream.u8(self.area)
		stream.u8(self.language)
		stream.u8(self.platform)
		stream.add(self.game_key)
		stream.string(self.message)
		stream.datetime(self.message_updated)
		stream.datetime(self.friended)
		stream.datetime(self.unk)


class FriendsProtocolV1:
	METHOD_UPDATE_PROFILE = 1
	METHOD_UPDATE_MII = 2
	METHOD_UPDATE_MII_LIST = 3
	METHOD_UPDATE_PLAYED_GAMES = 4
	METHOD_UPDATE_PREFERENCE = 5
	METHOD_GET_FRIEND_MII = 6
	METHOD_GET_FRIEND_MII_LIST = 7
	METHOD_GET_FRIEND_RELATIONSHIPS = 10
	METHOD_ADD_FRIEND_BY_PRINCIPAL_ID = 11
	METHOD_GET_ALL_FRIENDS = 15
	METHOD_SYNC_FRIEND = 17
	METHOD_UPDATE_PRESENCE = 18
	METHOD_UPDATE_FAVORITE_GAME_KEY = 19
	METHOD_UPDATE_COMMENT = 20
	METHOD_GET_FRIEND_PRESENCE = 22
	METHOD_GET_FRIEND_PICTURE = 24
	METHOD_GET_FRIEND_PERSISTENT_INFO = 25
	METHOD_SEND_INVITATION = 26
	
	PROTOCOL_ID = 0x65


class FriendsClientV1(FriendsProtocolV1):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def update_profile(self, profile_data):
		logger.info("FriendsClientV1.update_profile()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(profile_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_PROFILE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_profile -> done")
	
	async def update_mii(self, mii):
		logger.info("FriendsClientV1.update_mii()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(mii)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_MII, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_mii -> done")
	
	async def update_mii_list(self, mii_list):
		logger.info("FriendsClientV1.update_mii_list()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(mii_list)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_MII_LIST, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_mii_list -> done")
	
	async def update_played_games(self, played_games):
		logger.info("FriendsClientV1.update_played_games()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(played_games, stream.add)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_PLAYED_GAMES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_played_games -> done")
	
	async def update_preference(self, unk1, unk2, unk3):
		logger.info("FriendsClientV1.update_preference()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.bool(unk1)
		stream.bool(unk2)
		stream.bool(unk3)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_PREFERENCE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_preference -> done")
	
	async def get_friend_mii(self, friends):
		logger.info("FriendsClientV1.get_friend_mii()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(friends, stream.add)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_FRIEND_MII, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		miis = stream.list(FriendMii)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.get_friend_mii -> done")
		return miis
	
	async def get_friend_mii_list(self, friends):
		logger.info("FriendsClientV1.get_friend_mii_list()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(friends, stream.add)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_FRIEND_MII_LIST, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		mii_lists = stream.list(FriendMiiList)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.get_friend_mii_list -> done")
		return mii_lists
	
	async def get_friend_relationships(self, unk):
		logger.info("FriendsClientV1.get_friend_relationships()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(unk, stream.u32)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_FRIEND_RELATIONSHIPS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		friend_relationships = stream.list(FriendRelationship)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.get_friend_relationships -> done")
		return friend_relationships
	
	async def add_friend_by_principal_id(self, unk, pid):
		logger.info("FriendsClientV1.add_friend_by_principal_id()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(unk)
		stream.u32(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_ADD_FRIEND_BY_PRINCIPAL_ID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		friend_relationship = stream.extract(FriendRelationship)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.add_friend_by_principal_id -> done")
		return friend_relationship
	
	async def get_all_friends(self):
		logger.info("FriendsClientV1.get_all_friends()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_ALL_FRIENDS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		friend_relationships = stream.list(FriendRelationship)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.get_all_friends -> done")
		return friend_relationships
	
	async def sync_friend(self, unk1, unk2, unk3):
		logger.info("FriendsClientV1.sync_friend()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(unk1)
		stream.list(unk2, stream.u32)
		stream.list(unk3, stream.u64)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_SYNC_FRIEND, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		friend_list = stream.list(FriendRelationship)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.sync_friend -> done")
		return friend_list
	
	async def update_presence(self, presence_info, unk):
		logger.info("FriendsClientV1.update_presence()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(presence_info)
		stream.bool(unk)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_PRESENCE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_presence -> done")
	
	async def update_favorite_game_key(self, game_key):
		logger.info("FriendsClientV1.update_favorite_game_key()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(game_key)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_FAVORITE_GAME_KEY, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_favorite_game_key -> done")
	
	async def update_comment(self, comment):
		logger.info("FriendsClientV1.update_comment()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(comment)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_COMMENT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.update_comment -> done")
	
	async def get_friend_presence(self, unk):
		logger.info("FriendsClientV1.get_friend_presence()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(unk, stream.u32)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_FRIEND_PRESENCE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		friend_presence_list = stream.list(FriendPresence)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.get_friend_presence -> done")
		return friend_presence_list
	
	async def get_friend_picture(self, unk):
		logger.info("FriendsClientV1.get_friend_picture()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(unk, stream.u32)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_FRIEND_PICTURE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		friend_pictures = stream.list(FriendPicture)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.get_friend_picture -> done")
		return friend_pictures
	
	async def get_friend_persistent_info(self, unk):
		logger.info("FriendsClientV1.get_friend_persistent_info()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(unk, stream.u32)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_FRIEND_PERSISTENT_INFO, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		persistent_infos = stream.list(FriendPersistentInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.get_friend_persistent_info -> done")
		return persistent_infos
	
	async def send_invitation(self, unk):
		logger.info("FriendsClientV1.send_invitation()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(unk, stream.u32)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_SEND_INVITATION, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("FriendsClientV1.send_invitation -> done")


class FriendsServerV1(FriendsProtocolV1):
	def __init__(self):
		self.methods = {
			self.METHOD_UPDATE_PROFILE: self.handle_update_profile,
			self.METHOD_UPDATE_MII: self.handle_update_mii,
			self.METHOD_UPDATE_MII_LIST: self.handle_update_mii_list,
			self.METHOD_UPDATE_PLAYED_GAMES: self.handle_update_played_games,
			self.METHOD_UPDATE_PREFERENCE: self.handle_update_preference,
			self.METHOD_GET_FRIEND_MII: self.handle_get_friend_mii,
			self.METHOD_GET_FRIEND_MII_LIST: self.handle_get_friend_mii_list,
			self.METHOD_GET_FRIEND_RELATIONSHIPS: self.handle_get_friend_relationships,
			self.METHOD_ADD_FRIEND_BY_PRINCIPAL_ID: self.handle_add_friend_by_principal_id,
			self.METHOD_GET_ALL_FRIENDS: self.handle_get_all_friends,
			self.METHOD_SYNC_FRIEND: self.handle_sync_friend,
			self.METHOD_UPDATE_PRESENCE: self.handle_update_presence,
			self.METHOD_UPDATE_FAVORITE_GAME_KEY: self.handle_update_favorite_game_key,
			self.METHOD_UPDATE_COMMENT: self.handle_update_comment,
			self.METHOD_GET_FRIEND_PRESENCE: self.handle_get_friend_presence,
			self.METHOD_GET_FRIEND_PICTURE: self.handle_get_friend_picture,
			self.METHOD_GET_FRIEND_PERSISTENT_INFO: self.handle_get_friend_persistent_info,
			self.METHOD_SEND_INVITATION: self.handle_send_invitation,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on FriendsServerV1: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_update_profile(self, client, input, output):
		logger.info("FriendsServerV1.update_profile()")
		#--- request ---
		profile_data = input.extract(MyProfile)
		await self.update_profile(client, profile_data)
	
	async def handle_update_mii(self, client, input, output):
		logger.info("FriendsServerV1.update_mii()")
		#--- request ---
		mii = input.extract(Mii)
		await self.update_mii(client, mii)
	
	async def handle_update_mii_list(self, client, input, output):
		logger.info("FriendsServerV1.update_mii_list()")
		#--- request ---
		mii_list = input.extract(MiiList)
		await self.update_mii_list(client, mii_list)
	
	async def handle_update_played_games(self, client, input, output):
		logger.info("FriendsServerV1.update_played_games()")
		#--- request ---
		played_games = input.list(PlayedGame)
		await self.update_played_games(client, played_games)
	
	async def handle_update_preference(self, client, input, output):
		logger.info("FriendsServerV1.update_preference()")
		#--- request ---
		unk1 = input.bool()
		unk2 = input.bool()
		unk3 = input.bool()
		await self.update_preference(client, unk1, unk2, unk3)
	
	async def handle_get_friend_mii(self, client, input, output):
		logger.info("FriendsServerV1.get_friend_mii()")
		#--- request ---
		friends = input.list(FriendMiiRequest)
		response = await self.get_friend_mii(client, friends)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_get_friend_mii_list(self, client, input, output):
		logger.info("FriendsServerV1.get_friend_mii_list()")
		#--- request ---
		friends = input.list(FriendMiiRequest)
		response = await self.get_friend_mii_list(client, friends)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_get_friend_relationships(self, client, input, output):
		logger.info("FriendsServerV1.get_friend_relationships()")
		#--- request ---
		unk = input.list(input.u32)
		response = await self.get_friend_relationships(client, unk)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_add_friend_by_principal_id(self, client, input, output):
		logger.info("FriendsServerV1.add_friend_by_principal_id()")
		#--- request ---
		unk = input.u64()
		pid = input.u32()
		response = await self.add_friend_by_principal_id(client, unk, pid)
		
		#--- response ---
		if not isinstance(response, FriendRelationship):
			raise RuntimeError("Expected FriendRelationship, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_all_friends(self, client, input, output):
		logger.info("FriendsServerV1.get_all_friends()")
		#--- request ---
		response = await self.get_all_friends(client)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_sync_friend(self, client, input, output):
		logger.info("FriendsServerV1.sync_friend()")
		#--- request ---
		unk1 = input.u64()
		unk2 = input.list(input.u32)
		unk3 = input.list(input.u64)
		response = await self.sync_friend(client, unk1, unk2, unk3)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_update_presence(self, client, input, output):
		logger.info("FriendsServerV1.update_presence()")
		#--- request ---
		presence_info = input.extract(NintendoPresence)
		unk = input.bool()
		await self.update_presence(client, presence_info, unk)
	
	async def handle_update_favorite_game_key(self, client, input, output):
		logger.info("FriendsServerV1.update_favorite_game_key()")
		#--- request ---
		game_key = input.extract(GameKey)
		await self.update_favorite_game_key(client, game_key)
	
	async def handle_update_comment(self, client, input, output):
		logger.info("FriendsServerV1.update_comment()")
		#--- request ---
		comment = input.string()
		await self.update_comment(client, comment)
	
	async def handle_get_friend_presence(self, client, input, output):
		logger.info("FriendsServerV1.get_friend_presence()")
		#--- request ---
		unk = input.list(input.u32)
		response = await self.get_friend_presence(client, unk)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_get_friend_picture(self, client, input, output):
		logger.info("FriendsServerV1.get_friend_picture()")
		#--- request ---
		unk = input.list(input.u32)
		response = await self.get_friend_picture(client, unk)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_get_friend_persistent_info(self, client, input, output):
		logger.info("FriendsServerV1.get_friend_persistent_info()")
		#--- request ---
		unk = input.list(input.u32)
		response = await self.get_friend_persistent_info(client, unk)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_send_invitation(self, client, input, output):
		logger.info("FriendsServerV1.send_invitation()")
		#--- request ---
		unk = input.list(input.u32)
		await self.send_invitation(client, unk)
	
	async def update_profile(self, *args):
		logger.warning("FriendsServerV1.update_profile not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_mii(self, *args):
		logger.warning("FriendsServerV1.update_mii not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_mii_list(self, *args):
		logger.warning("FriendsServerV1.update_mii_list not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_played_games(self, *args):
		logger.warning("FriendsServerV1.update_played_games not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_preference(self, *args):
		logger.warning("FriendsServerV1.update_preference not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_friend_mii(self, *args):
		logger.warning("FriendsServerV1.get_friend_mii not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_friend_mii_list(self, *args):
		logger.warning("FriendsServerV1.get_friend_mii_list not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_friend_relationships(self, *args):
		logger.warning("FriendsServerV1.get_friend_relationships not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def add_friend_by_principal_id(self, *args):
		logger.warning("FriendsServerV1.add_friend_by_principal_id not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_all_friends(self, *args):
		logger.warning("FriendsServerV1.get_all_friends not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def sync_friend(self, *args):
		logger.warning("FriendsServerV1.sync_friend not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_presence(self, *args):
		logger.warning("FriendsServerV1.update_presence not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_favorite_game_key(self, *args):
		logger.warning("FriendsServerV1.update_favorite_game_key not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_comment(self, *args):
		logger.warning("FriendsServerV1.update_comment not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_friend_presence(self, *args):
		logger.warning("FriendsServerV1.get_friend_presence not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_friend_picture(self, *args):
		logger.warning("FriendsServerV1.get_friend_picture not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_friend_persistent_info(self, *args):
		logger.warning("FriendsServerV1.get_friend_persistent_info not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def send_invitation(self, *args):
		logger.warning("FriendsServerV1.send_invitation not implemented")
		raise common.RMCError("Core::NotImplemented")





from anynet import tls, http

import logging
logger = logging.getLogger(__name__)


USER_AGENT = {
	 900: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 901: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 910: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 920: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	1000: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1001: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1002: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1003: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1004: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1010: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1011: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1020: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1100: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1101: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1200: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1201: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1202: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1203: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1210: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1300: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.3.0.0; Add-on 13.3.0.0)",
	1310: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1320: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1321: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1400: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1410: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1411: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1412: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
}

LATEST_VERSION = 1412


class BAASError(Exception):
	def __init__(self, error):
		self.status = error["status"]
		self.name = error["errorCode"]
		self.title = error["title"]
		self.detail = error["detail"]
		self.instance = error["instance"]
		self.type = error["type"]


class BAASClient:
	def __init__(self):
		self.url = "e0d67c509fb203858ebcb2fe3f88c2aa.baas.nintendo.com"
		self.user_agent = USER_AGENT[LATEST_VERSION]
		self.power_state = "FA"
		
		self.context = tls.TLSContext()
	
	def set_url(self, url): self.url = url
	def set_user_agent(self, user_agent): self.user_agent = user_agent
	def set_power_state(self, state): self.power_state = state
	def set_context(self, context): self.context = context
	
	def set_system_version(self, version):
		if version not in USER_AGENT:
			raise ValueError("Unknown system version")
		self.user_agent = USER_AGENT[version]
		
	async def request(self, req, token, use_power_state):
		req.headers["Host"] = self.url
		req.headers["User-Agent"] = self.user_agent
		req.headers["Accept"] = "*/*"
		if token:
			req.headers["Authorization"] = "Bearer " + token
		if use_power_state:
			req.headers["X-Nintendo-PowerState"] = self.power_state
		if req.method == "POST":
			req.headers["Content-Length"] = 0
			req.headers["Content-Type"] = "application/x-www-form-urlencoded"
		
		response = await http.request(self.url, req, self.context)
		if response.error():
			logger.warning("BAAS request returned error: %s" %response.json)
			raise BAASError(response.json)
		return response
		
	async def authenticate(self, device_token):
		req = http.HTTPRequest.post("/1.0.0/application/token")
		req.form = {
			"grantType": "public_client",
			"assertion": device_token
		}

		response = await self.request(req, None, True)
		return response.json
	
	async def login(self, id, password, access_token, app_token=None, skip_verification=False):
		req = http.HTTPRequest.post("/1.0.0/login")
		req.form = {
			"id": "%016x" %id,
			"password": password
		}
		
		if app_token:
			req.form["appAuthNToken"] = app_token
		if skip_verification:
			req.form["skipOp2Verification"] = "1"
			
		response = await self.request(req, access_token, True)
		return response.json
	
	async def register(self, access_token):
		req = http.HTTPRequest.post("/1.0.0/users")
		
		response = await self.request(req, access_token, False)
		return response.json




from anynet import streams


class StreamOut(streams.StreamOut):
	def __init__(self, settings):
		super().__init__(">")
		self.settings = settings
		
	def pid(self, value):
		if self.settings["common.pid_size"] == 8:
			self.u64(value)
		else:
			self.u32(value)
		
	def add(self, inst, *args):
		inst.encode(self, *args)
		
		
class StreamIn(streams.StreamIn):
	def __init__(self, data, settings):
		super().__init__(data, ">")
		self.settings = settings
		
	def pid(self):
		if self.settings["common.pid_size"] == 8:
			return self.u64()
		return self.u32()
		
	def extract(self, cls, *args):
		inst = cls()
		inst.decode(self, *args)
		return inst




class Random:
	def __init__(self, *param):
		if len(param) == 1: self.set_seed(param[0])
		elif len(param) == 4: self.set_state(*param)
		else:
			raise TypeError("Random.__init__ takes either 1 or 4 arguments")
			
	def set_seed(self, seed):
		multiplier = 0x6C078965
		
		temp = seed
		self.state = []
		for i in range(1, 5):
			temp ^= temp >> 30
			temp = (temp * multiplier + i) & 0xFFFFFFFF
			self.state.append(temp)
			
	def set_state(self, s0, s1, s2, s3):
		self.state = [s0, s1, s2, s3]

	def u32(self):
		temp = self.state[0]
		temp = (temp ^ (temp << 11)) & 0xFFFFFFFF
		temp ^= temp >> 8
		temp ^= self.state[3]
		temp ^= self.state[3] >> 19
		self.state[0] = self.state[1]
		self.state[1] = self.state[2]
		self.state[2] = self.state[3]
		self.state[3] = temp
		return temp
	
	def uint(self, max):
		return (self.u32() * max) >> 32




from nintendo.nex import common
import socket


class Range:
	def __init__(self, min=0, max=10):
		self.min = min
		self.max = max
	
	def __repr__(self):
		return "<Range: %i - %i>" %(self.min, self.max)
	
	def __contains__(self, value):
		return self.min <= value <= self.max


class ResultRange:
	def __init__(self, offset=0, size=10):
		self.offset = offset
		self.size = size
		
	def encode(self, stream):
		stream.u32(self.offset)
		stream.u32(self.size)
		
	def decode(self, stream):
		self.offset = stream.u32()
		self.size = stream.u32()
		
		
class InetAddress:
	EMPTY = 2
	IPV4 = 6
	IPV6 = 18
	
	def __init__(self, host="0.0.0.0", port=0):
		self.host = host
		self.port = port
		
	def __bool__(self):
		return self.host != "0.0.0.0" or self.port == 0
		
	def get_type(self):
		if self.host == "0.0.0.0":
			return InetAddress.EMPTY
		return InetAddress.IPV4
		
	def encode(self, stream, type=IPV4):
		if type == InetAddress.IPV4:
			stream.write(socket.inet_aton(self.host))
		elif type == InetAddress.IPV6:
			stream.write(socket.inet_aton(self.host))
			stream.pad(12)
		elif type != InetAddress.EMPTY:
			raise ValueError("Invalid InetAddress type: %i" %type)
		stream.u16(self.port)
		
	def decode(self, stream, type=IPV4):
		if type == InetAddress.EMPTY:
			self.host = "0.0.0.0"
		elif type == InetAddress.IPV4:
			self.host = socket.inet_ntoa(stream.read(4))
		elif type == InetAddress.IPV6:
			self.host = socket.inet_ntoa(stream.read(4))
			if stream.read(12) != bytes(12):
				raise ValueError("IPv6 addresses are not supported")
		else:
			raise ValueError("Invalid InetAddress type: %i" %type)
		self.port = stream.u16()


class StationAddress:
	def __init__(self, host="0.0.0.0", port=0):
		self.inet = InetAddress(host, port)
		self.extension_id = 0
		
	def encode(self, stream, type=InetAddress.IPV4):
		stream.add(self.inet, type)
		if stream.settings["pia.version"] < 500:
			stream.u16(self.extension_id)
			
	def decode(self, stream, type=InetAddress.IPV4):
		self.inet = stream.extract(InetAddress, type)
		if stream.settings["pia.version"] < 500:
			self.extension_id = stream.u16()
		
		
class StationLocation:
	url_types = {
		None: 0,
		"prudp": 1,
		"prudps": 2,
		"udp": 3
	}
	
	url_schemes = {
		0: None,
		1: "prudp",
		2: "prudps",
		3: "udp"
	}
	
	def __init__(self):
		self.public = StationAddress()
		self.local = StationAddress()
		self.pid = 0
		self.cid = 0
		self.rvcid = 0
		self.scheme = 0
		self.sid = 0
		self.stream_type = 0
		self.natm = 0
		self.natf = 0
		self.type = 3
		self.probeinit = 0
		self.relay = InetAddress()
		
	def copy(self):
		copy = StationLocation()
		copy.set_station_url(self.get_station_url())
		return copy
		
	def set_station_url(self, url):
		self.public = StationAddress()
		self.local = StationAddress()
		
		if url["type"] == 0:
			self.local.inet = InetAddress(url["address"], url["port"])
		else:
			self.public.inet = InetAddress(url["address"], url["port"])
		
		self.pid = url["PID"]
		self.cid = url["CID"]
		self.rvcid = url["RVCID"]
		self.scheme = self.url_types[url.scheme()]
		self.sid = url["sid"]
		self.stream_type = url["stream"]
		self.natm = url["natm"]
		self.natf = url["natf"]
		self.type = url["type"]
		self.probeinit = url["probeinit"]
		
		self.relay = InetAddress(url["Rsa"], url["Rsp"])
		
	def get_station_url(self):
		url = common.StationURL(self.url_schemes[self.scheme])
		if self.type == 0:
			url["address"] = self.local.inet.host
			url["port"] = self.local.inet.port
		else:
			url["address"] = self.public.inet.host
			url["port"] = self.public.inet.port
		url["PID"] = self.pid
		url["CID"] = self.cid
		url["RVCID"] = self.rvcid
		url["sid"] = self.sid
		url["stream"] = self.stream_type
		url["natm"] = self.natm
		url["natf"] = self.natf
		url["type"] = self.type
		url["probeinit"] = self.probeinit
		
		url["Rsa"] = self.relay.host
		url["Rsp"] = self.relay.port
		
		return url
		
	def encode(self, stream):
		if stream.settings["pia.version"] < 510:
			stream.add(self.public)
		else:
			type_public = InetAddress.IPV4
			type_local = InetAddress.IPV4
			if stream.settings["pia.version"] >= 511:
				type_public = self.public.inet.get_type()
				type_local = self.local.inet.get_type()
				stream.u8(type_public)
				stream.u8(type_local)
			stream.add(self.public, type_public)
			stream.add(self.local, type_local)
			stream.add(self.relay)
		stream.pid(self.pid)
		stream.u32(self.cid)
		stream.u32(self.rvcid)
		if stream.settings["pia.version"] < 510:
			stream.u8(self.scheme)
			stream.u8(self.sid)
			stream.u8(self.stream_type)
			stream.u8(self.natm)
			stream.u8(self.natf)
			stream.u8(self.type)
			stream.u8(self.probeinit)
			if stream.settings["pia.version"] >= 500:
				stream.add(self.relay)
		else:
			stream.u8((self.natm << 2) | self.natf)
			stream.u8(self.type)
			stream.u8(self.probeinit)
			stream.bool(self.type == 0)
		
	def decode(self, stream):
		if stream.settings["pia.version"] < 510:
			self.public = stream.extract(StationAddress)
		else:
			if stream.settings["pia.version"] >= 511:
				type_public = stream.u8()
				type_local = stream.u8()
			else:
				type_public = InetAddress.IPV4
				type_local = InetAddress.IPV4
			self.public = stream.extract(StationAddress, type_public)
			self.local = stream.extract(StationAddress, type_local)
			self.relay = stream.extract(InetAddress)
		self.pid = stream.pid()
		self.cid = stream.u32()
		self.rvcid = stream.u32()
		if stream.settings["pia.version"] < 510:
			self.scheme = stream.u8()
			self.sid = stream.u8()
			self.stream_type = stream.u8()
			self.natm = stream.u8()
			self.natf = stream.u8()
			self.type = stream.u8()
			self.probeinit = stream.u8()
			if stream.settings["pia.version"] >= 500:
				self.relay = stream.extract(InetAddress)
		else:
			natinfo = stream.u8()
			self.natm = natinfo >> 2
			self.natf = natinfo & 3
			self.type = stream.u8()
			self.probeinit = stream.u8()
			stream.skip(1)




from Crypto.Cipher import AES
from nintendo.pia import streams, types
from anynet import udp, util
import contextlib
import itertools
import secrets
import hashlib
import socket
import struct
import anyio
import hmac

import logging
logger = logging.getLogger(__name__)


nonce_counter = itertools.count()


def encrypt_key(kek, key):
	aes = AES.new(kek, AES.MODE_ECB)
	return aes.encrypt(key)

def generate_nonce(nonce_id):
	addr = util.broadcast_address()
	return struct.pack(">IQ", util.ip_to_hex(addr), nonce_id)

def generate_challenge_reply(game_key, challenge_key, challenge_data, nonce):
	key = hmac.digest(game_key, challenge_key, hashlib.sha256)[:16]
	data = hmac.digest(game_key, challenge_data, hashlib.sha256)[:16]
	
	aes = AES.new(key, AES.MODE_GCM, nonce=nonce)
	ciphertext, tag = aes.encrypt_and_digest(data)
	return tag + ciphertext

def make_browse_request(settings, search_criteria, game_key, challenge_key, challenge_data):
	stream = streams.StreamOut(settings)
	stream.add(search_criteria)
	buffer = stream.get()
	
	stream = streams.StreamOut(settings)
	stream.u8(0) #Packet type
	stream.u32(len(buffer))
	stream.write(buffer)
	
	if settings["pia.lan_version"] != 0:
		nonce_id = next(nonce_counter)
		stream.u8(settings["pia.lan_version"])
		stream.bool(game_key is not None)
		stream.u64(nonce_id)
		stream.write(challenge_key)
		
		if game_key is not None:
			key = encrypt_key(game_key, challenge_key)
			nonce = generate_nonce(nonce_id)
			
			aes = AES.new(key, AES.MODE_GCM, nonce=nonce)
			ciphertext, tag = aes.encrypt_and_digest(challenge_data)
			challenge = tag + ciphertext
		else:
			challenge = secrets.token_bytes(16 + 256)
		stream.write(challenge)
	return stream.get()

def verify_challenge_reply(stream, game_key, challenge_key, challenge_data):
	if stream.settings["pia.lan_version"] == 0:
		return True
	
	version = stream.u8()
	expected = stream.settings["pia.lan_version"]
	if version != expected:
		logger.warning("Unexpected version in challenge reply header (%i != %i)", version, expected)
		return False
		
	enabled = stream.bool()
	if enabled and game_key is None:
		logger.warning("Received encrypted challenge reply but no game key was provided")
		return False
	
	if not enabled:
		stream.skip(8 + 16 + 32)
		return True
	
	nonce_id = stream.u64()
	key = stream.read(16) + challenge_key
	reply = stream.read(32)
	
	nonce = generate_nonce(nonce_id)
	expected = generate_challenge_reply(game_key, key, challenge_data, nonce)
	
	if reply != expected:
		logger.warning("Incorrect challenge reply received")
		return False
	return True

def parse_browse_reply(settings, data, game_key, challenge_key, challenge_data):
	stream = streams.StreamIn(data, settings)
	if stream.u8() != 1:
		logger.warning("Received packet that is not a browse reply")
		return None
	
	size = stream.u32()
	data = stream.read(size)
	
	if not verify_challenge_reply(stream, game_key, challenge_key, challenge_data):
		return None
	
	if not stream.eof():
		logger.warning("Browse reply is bigger than expected")
		return None
	
	stream = streams.StreamIn(data, settings)
	try:
		session_info = stream.extract(LanSessionInfo)
	except Exception as e:
		logger.warning("Failed to parse LanSessionInfo: %s", e)
		return None
	
	if not stream.eof():
		logger.warning("LanSessionInfo has unexpected size")
		return None
	
	return session_info
	

def default(value, default):
	if value is None:
		return default
	return value

class LanSessionSearchCriteria:
	def __init__(self):
		self.reset()
		
	def reset(self):
		self.min_participants = None
		self.max_participants = None
		self.opened_only = None
		self.vacant_only = None
		self.result_range = types.ResultRange()
		self.game_mode = None
		self.session_type = None
		self.attributes = [None] * 6
		
	def check(self, info):
		if self.min_participants is not None:
			if info.min_participants not in self.min_participants:
				logger.info("Search criteria miss: min_participants")
				return False
		if self.max_participants is not None:
			if info.max_participants not in self.max_participants:
				logger.info("Search criteria miss: max_participants")
				return False
		if self.opened_only and not info.is_opened:
			logger.info("Search criteria miss: session is not opened")
			return False
		if self.vacant_only and info.num_participants >= info.max_participants:
			logger.info("Search criteria miss: session is full")
			return False
		if self.game_mode is not None and self.game_mode != info.game_mode:
			logger.info("Search criteria miss: different game mode")
			return False
		if self.session_type is not None and self.session_type != info.session_type:
			logger.info("Search criteria miss: different session type")
			return False
		for i in range(6):
			if self.attributes[i] is not None:
				if info.attributes[i] not in self.attributes[i]:
					logger.info("Search criteria miss: different attributes")
					return False
		return True
		
	def calc_flags(self):
		flags = 0
		if self.min_participants is not None: flags |= 1
		if self.max_participants is not None: flags |= 2
		if self.opened_only is not None: flags |= 4
		if self.vacant_only is not None: flags |= 8
		if self.game_mode is not None: flags |= 16
		if self.session_type is not None: flags |= 32
		for i in range(6):
			if self.attributes[i] is not None:
				flags |= 64 << i
		return flags
	
	def encode(self, stream):
		if self.min_participants:
			stream.u16(self.min_participants.max)
			stream.u16(self.min_participants.min)
		else:
			stream.u32(0)
			
		if self.max_participants:
			stream.u16(self.max_participants.max)
			stream.u16(self.max_participants.min)
		else:
			stream.u32(0)
		
		stream.bool(default(self.opened_only, False))
		stream.bool(default(self.vacant_only, False))
		stream.add(self.result_range)
		stream.u32(default(self.game_mode, 0))
		stream.u32(default(self.session_type, 0))
		
		for attrib in self.attributes:
			if isinstance(attrib, list):
				if len(attrib) > 20:
					raise ValueError("Attribute list may only contain up to 20 values")
				stream.repeat(attrib + [0] * (20 - len(attrib)), stream.u32)
			else:
				stream.repeat([0] * 20, stream.u32)
				
		for attrib in self.attributes:
			if isinstance(attrib, list):
				stream.u8(len(attrib))
			else:
				stream.u8(0)
				
		for attrib in self.attributes:
			if isinstance(attrib, types.Range):
				stream.u32(attrib.min)
			else:
				stream.u32(0)
		
		for attrib in self.attributes:
			if isinstance(attrib, types.Range):
				stream.u32(attrib.max)
			else:
				stream.u32(0)
				
		for attrib in self.attributes:
			if isinstance(attrib, types.Range):
				stream.bool(True)
			else:
				stream.bool(False)
				
		stream.u32(self.calc_flags())
		
	def decode(self, stream):
		self.reset()
		
		min_participants = types.Range()
		max_participants = types.Range()
		
		min_participants.max = stream.u16()
		min_participants.min = stream.u16()
		max_participants.max = stream.u16()
		max_participants.min = stream.u16()
		opened_only = stream.bool()
		vacant_only = stream.bool()
		result_range = stream.extract(types.ResultRange)
		game_mode = stream.u32()
		session_type = stream.u32()
		
		attrib_values = stream.repeat(
			lambda: stream.repeat(stream.u32, 20), 6
		)
		attrib_sizes = stream.repeat(stream.u8, 6)
		attrib_range_min = stream.repeat(stream.u32, 6)
		attrib_range_max = stream.repeat(stream.u32, 6)
		attrib_mode = stream.repeat(stream.bool, 6)
		
		flags = stream.u32()
		if flags & 1: self.min_participants = min_participants
		if flags & 2: self.max_participants = max_participants
		if flags & 4: self.opened_only = opened_only
		if flags & 8: self.vacant_only = vacant_only
		self.result_range = result_range
		if flags & 16: self.game_mode = game_mode
		if flags & 32: self.session_type = session_type
		
		for i in range(6):
			if flags & (64 << i):
				if attrib_mode[i]:
					self.attributes[i] = types.Range(attrib_range_min[i], attrib_range_max[i])
				else:
					self.attributes[i] = attrib_values[i][:attrib_sizes[i]]
					

class LanStationInfo:

	NONE = 0
	UTF8 = 1
	UTF16 = 2
	
	HOST = 1
	PLAYER = 2
	
	def __init__(self):
		self.role = 0
		self.encoding = self.UTF8
		self.username = ""
		self.id = 0
		
	def decode(self, stream):
		self.role = stream.u8()
		self.encoding = stream.u8()
		
		username = stream.read(40)
		if self.encoding == self.NONE:
			username = ""
		elif self.encoding == self.UTF8:
			username = username.decode("utf8")
		elif self.encoding == self.UTF16:
			username = username.decode("utf16")
		else:
			raise ValueError("Invalid username encoding: %i" %self.encoding)
		self.username = username.split("\0")[0]
		
		self.id = stream.u64()
		
	def encode(self, stream):
		stream.u8(self.role)
		stream.u8(self.encoding)
		
		if self.encoding == self.UTF8:
			username = self.username.encode("utf8")
		else:
			username = self.username.encode("utf16")
		username = username.ljust(40, b"\0")
		stream.write(username)
		
		stream.u64(self.id)


class LanSessionInfo:
	def __init__(self):
		self.game_mode = None
		self.session_id = None
		self.attributes = [0] * 6
		self.num_participants = 0
		self.min_participants = None
		self.max_participants = None
		self.system_version = None
		self.application_version = None
		self.session_type = None
		self.application_data = b""
		self.is_opened = None
		self.host_location = types.StationLocation()
		self.stations = [LanStationInfo() for i in range(16)]
		self.session_param = bytes(32)
	
	def get_session_key(self, key):
		param = self.session_param
		param = param[:31] + bytes([(param[31] + 1) & 0xFF])
		return hmac.digest(key, param, hashlib.sha256)
		
	def encode(self, stream):
		if len(self.attributes) != 6: raise ValueError("Attributes must contain exactly 6 integers")
		if len(self.stations) != 16:
			raise ValueError("Session must contain exactly 16 station entries")
		
		stream.u32(self.game_mode)
		stream.u32(self.session_id)
		stream.repeat(self.attributes, stream.u32)
		stream.u16(self.num_participants)
		stream.u16(self.min_participants)
		stream.u16(self.max_participants)
		if stream.settings["pia.version"] < 503:
			stream.u32(self.session_type)
		else:
			stream.u8(self.system_version)
			stream.u8(self.application_version)
			stream.u16(self.session_type)
		stream.write(self.application_data.ljust(0x180, b"\0"))
		stream.u32(len(self.application_data))
		stream.bool(self.is_opened)
		if stream.settings["pia.version"] < 510:
			stream.add(self.host_location)
		else:
			type = types.InetAddress.IPV4
			if stream.settings["pia.lan_version"] == 2:
				type = types.InetAddress.IPV6
			stream.add(self.host_location.local, type)
			stream.pid(self.host_location.pid)
			stream.u32(self.host_location.cid)
			stream.u32(self.host_location.rvcid)
		stream.repeat(self.stations, stream.add)
		if stream.settings["pia.lan_version"] >= 1:
			stream.write(self.session_param)
		
	def decode(self, stream):
		self.game_mode = stream.u32()
		self.session_id = stream.u32()
		self.attributes = stream.repeat(stream.u32, 6)
		self.num_participants = stream.u16()
		self.min_participants = stream.u16()
		self.max_participants = stream.u16()
		if stream.settings["pia.version"] < 503:
			self.session_type = stream.u32()
		else:
			self.system_version = stream.u8()
			self.application_version = stream.u8()
			self.session_type = stream.u16()
		self.application_data = stream.read(0x180)
		self.application_data = self.application_data[:stream.u32()]
		self.is_opened = stream.bool()
		if stream.settings["pia.version"] < 510:
			self.host_location = stream.extract(types.StationLocation)
		else:
			type = types.InetAddress.IPV4
			if stream.settings["pia.lan_version"] == 2:
				type = types.InetAddress.IPV6
			self.host_location = types.StationLocation()
			self.host_location.local = stream.extract(types.StationAddress, type)
			self.host_location.pid = stream.pid()
			self.host_location.cid = stream.u32()
			self.host_location.rvcid = stream.u32()
		self.players = stream.repeat(
			lambda: stream.extract(LanStationInfo), 16
		)
		if stream.settings["pia.lan_version"] >= 1:
			self.session_param = stream.read(32)


class LanServer:
	def __init__(self, settings, handler, key, sock, group):
		self.settings = settings
		self.handler = handler
		self.game_key = key
		self.sock = sock
		self.group = group
	
	def start(self):
		self.group.start_soon(self.process)
	
	async def process(self):
		while True:
			data, addr = await self.sock.recv()
			await self.process_request(data, addr)
	
	async def process_request(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u8() != 0:
			logger.debug("Message is not a browse request")
			return
		
		size = stream.u32()
		end = stream.tell() + size
		
		criteria = stream.extract(LanSessionSearchCriteria)
		if stream.tell() != end:
			logger.warning("LanSessionSearchCriteria has unexpected size")
			return
		
		challenge_response = self.process_challenge(stream)
		if challenge_response is None:
			return
		
		if not stream.eof():
			logger.warning("Browse request is bigger than expected")
			return
		
		logger.info("Received browse request")
		
		sessions = []
		for session in self.handler():
			if criteria.check(session):
				sessions.append(session)
		
		logger.info("%i sessions match search criteria", len(sessions))
		for session in sessions:
			data = self.make_browse_reply(session, challenge_response)
			await self.sock.send(data, addr)
	
	def make_browse_reply(self, session_info, challenge):
		stream = streams.StreamOut(self.settings)
		stream.add(session_info)
		buffer = stream.get()
		
		stream = streams.StreamOut(self.settings)
		stream.u8(1) #Packet type
		stream.u32(len(buffer))
		stream.write(buffer)
		
		if self.settings["pia.lan_version"] != 0:
			stream.u8(self.settings["pia.lan_version"])
			stream.bool(self.game_key is not None)
			stream.write(challenge)
		
		return stream.get()
	
	def process_challenge(self, stream):
		if self.settings["pia.lan_version"] == 0:
			return b""
		
		version = stream.u8()
		if version != self.settings["pia.lan_version"]:
			logger.warning("Browse request has unexpected version number")
			return None
		
		crypto = stream.bool()
		if crypto and self.game_key is None:
			logger.warning("Received crypto challenge but no game key was provided")
			return None
		
		nonce_id = stream.u64()
		key = stream.read(16)
		
		tag = stream.read(16)
		challenge = stream.read(256)
		
		if crypto:
			challenge_key = encrypt_key(self.game_key, key)
			nonce = generate_nonce(nonce_id)
			
			aes = AES.new(challenge_key, AES.MODE_GCM, nonce=nonce)
			try:
				challenge = aes.decrypt_and_verify(challenge, tag)
			except ValueError:
				logger.warning("Challenge has incorrect authentication tag")
				return None
			
			new_key = secrets.token_bytes(16)
			nonce_id = next(nonce_counter)
			nonce = generate_nonce(nonce_id)
			
			response = struct.pack(">Q", nonce_id) + new_key
			response += generate_challenge_reply(self.game_key, new_key + key, challenge, nonce)
		else:
			response = secrets.token_bytes(16 + 16 + 16)
		return response


async def browse(settings, search_criteria, key=None, timeout=1, max=0):
	challenge_key = secrets.token_bytes(16)
	challenge_data = secrets.token_bytes(256)
	async with udp.bind(broadcast=True) as sock:
		request = make_browse_request(settings, search_criteria, key, challenge_key, challenge_data)
		await sock.send(request, (util.broadcast_address(), 30000))
		
		sessions = []
		ids = []
		with anyio.move_on_after(timeout):
			while max == 0 or len(sessions) < max:
				data, addr = await sock.recv()
				session = parse_browse_reply(settings, data, key, challenge_key, challenge_data)
				if session and session.session_id not in ids:
					ids.append(session.session_id)
					sessions.append(session)
	return sessions

@contextlib.asynccontextmanager
async def serve(settings, handler, key=None):
	async with udp.bind("0.0.0.0", 30000) as sock:
		async with util.create_task_group() as group:
			server = LanServer(settings, handler, key, sock, group)
			server.start()
			yield server




from nintendo.pia import streams
from Crypto.Cipher import AES
import random
import struct
import time

import logging
logger = logging.getLogger(__name__)


def pad(data):
	return data + b"\xFF" * ((16 - len(data) % 16) % 16)


class Packet:
	def __init__(self):
		self.encrypted = False
		self.connection_id = 0
		self.destination_id = 0
		self.source_id = 0
		self.sequence_id = 0
		self.session_timer = 0
		self.rtt_timer = 0
		self.nonce = 0
		self.payload = b""
		self.footer = b""


class PacketEncoderV1:
	def __init__(self, network, settings):
		self.settings = settings
	
	def decode(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u32() != 0x32AB9864:
			raise ValueError("Packet has wrong magic number")
		
		encryption = stream.u8()
		if encryption not in [1, 2]:
			raise ValueError("Packet has unexpected encryption mode")
		
		packet = Packet()
		packet.encrypted = encryption == 2
		packet.connection_id = stream.u8()
		packet.sequence_id = stream.u16()
		packet.session_timer = stream.u16()
		packet.rtt_timer = stream.u16()
		packet.payload = stream.read(stream.available() - 16)
		
		signature = stream.read(16)
		if hmac.digest(self.settings["crypto.session_key"], data[:-16], "md5") != signature:
			raise ValueError("Signature is invalid")
		
		if packet.encrypted:
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_ECB)
			packet.payload = aes.decrypt(packet.payload)
		
		return packet
	
	def encode(self, packet, addr):
		stream = streams.StreamOut(self.settings)
		
		payload = packet.payload
		if packet.encrypted:
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_ECB)
			payload = aes.encrypt(pad(packet.payload))
		
		stream.u32(0x32AB9864)
		stream.u8(2 if packet.encrypted else 1)
		stream.u8(packet.connection_id)
		stream.u16(packet.sequence_id)
		stream.u16(packet.session_timer)
		stream.u16(packet.rtt_timer)
		stream.write(payload)
		stream.write(hmac.digest(self.settings["crypto.session_key"], stream.get(), "md5"))
		return stream.get()


class PacketEncoderV2:
	def __init__(self, network, settings):
		self.network = network
		self.settings = settings
	
	def decode(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u32() != 0x32AB9864:
			raise ValueError("Packet has wrong magic number")
		
		encryption = stream.u8()
		if encryption not in [1, 2]:
			raise ValueError("Packet has unexpected encryption mode")
		
		packet = Packet()
		packet.encrypted = encryption == 2
		packet.connection_id = stream.u8()
		packet.sequence_id = stream.u16()
		packet.session_timer = stream.u16()
		packet.rtt_timer = stream.u16()
		packet.nonce = stream.u64()
		signature = stream.read(16)
		packet.payload = stream.readall()
		
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			packet.payload = aes.decrypt_and_verify(packet.payload, signature)
		
		return packet
	
	def encode(self, packet, addr):
		signature = bytes(16)
		payload = packet.payload
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			payload, signature = aes.encrypt_and_digest(packet.payload)
		
		stream = streams.StreamOut(self.settings)
		stream.u32(0x32AB9864)
		stream.u8(2 if packet.encrypted else 1)
		stream.u8(packet.connection_id)
		stream.u16(packet.sequence_id)
		stream.u16(packet.session_timer)
		stream.u16(packet.rtt_timer)
		stream.u64(packet.nonce)
		stream.write(signature)
		stream.write(payload)
		return stream.get()


class PacketEncoderV3:
	def __init__(self, network, settings):
		self.network = network
		self.settings = settings
	
	def decode(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u32() != 0x32AB9864:
			raise ValueError("Packet has wrong magic number")
		
		encryption = stream.u8()
		if encryption & 0x7F != 3:
			raise ValueError("Packet has unexpected version number")
		
		packet = Packet()
		packet.encrypted = bool(encryption & 0x80)
		packet.connection_id = stream.u8()
		packet.sequence_id = stream.u16()
		packet.nonce = stream.u64()
		signature = stream.read(16)
		packet.payload = stream.readall()
		
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			packet.payload = aes.decrypt_and_verify(packet.payload, signature)
		
		return packet
	
	def encode(self, packet, addr):
		signature = bytes(16)
		payload = packet.payload
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			payload, signature = aes.encrypt_and_digest(packet.payload)
		
		stream = streams.StreamOut(self.settings)
		stream.u32(0x32AB9864)
		stream.u8((packet.encrypted << 7) | 3)
		stream.u8(packet.connection_id)
		stream.u16(packet.sequence_id)
		stream.u64(packet.nonce)
		stream.write(signature)
		stream.write(payload)
		return stream.get()


class PacketEncoderV4:
	def __init__(self, network, settings):
		self.network = network
		self.settings = settings
	
	def decode(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u32() != 0x32AB9864:
			raise ValueError("Packet has wrong magic number")
		
		encryption = stream.u8()
		if encryption & 0x7F != 4:
			raise ValueError("Packet has unexpected version number")
		
		packet = Packet()
		packet.encrypted = bool(encryption & 0x80)
		packet.connection_id = stream.u8()
		packet.sequence_id = stream.u16()
		packet.nonce = stream.u64()
		signature = stream.read(16)
		packet.payload = stream.readall()
		
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			packet.payload = aes.decrypt_and_verify(packet.payload, signature)
		
		return packet
	
	def encode(self, packet, addr):
		signature = bytes(16)
		payload = packet.payload
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			payload, signature = aes.encrypt_and_digest(packet.payload)
		
		stream = streams.StreamOut(self.settings)
		stream.u32(0x32AB9864)
		stream.u8((packet.encrypted << 7) | 4)
		stream.u8(packet.connection_id)
		stream.u16(packet.sequence_id)
		stream.u64(packet.nonce)
		stream.write(signature)
		stream.write(payload)
		return stream.get()


class PacketEncoderV5:
	def __init__(self, network, settings):
		self.network = network
		self.settings = settings
	
	def decode(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u32() != 0x32AB9864:
			raise ValueError("Packet has wrong magic number")
		
		encryption = stream.u8()
		if encryption & 0x7F != 5:
			raise ValueError("Packet has unexpected version number")
		
		packet = Packet()
		packet.encrypted = bool(encryption & 0x80)
		packet.connection_id = stream.u8()
		packet.sequence_id = stream.u16()
		packet.nonce = stream.u64()
		signature = stream.read(8)
		packet.payload = stream.readall()
		
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			packet.payload = aes.decrypt_and_verify(packet.payload, signature)
		
		return packet
	
	def encode(self, packet, addr):
		signature = bytes(16)
		payload = packet.payload
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			payload, signature = aes.encrypt_and_digest(packet.payload)
		
		stream = streams.StreamOut(self.settings)
		stream.u32(0x32AB9864)
		stream.u8((packet.encrypted << 7) | 5)
		stream.u8(packet.connection_id)
		stream.u16(packet.sequence_id)
		stream.u64(packet.nonce)
		stream.write(signature)
		stream.write(payload)
		return stream.get()


class PacketEncoderV9:
	def __init__(self, network, settings):
		self.network = network
		self.settings = settings
	
	def decode(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u32() != 0x32AB9864:
			raise ValueError("Packet has wrong magic number")
		
		encryption = stream.u8()
		if encryption & 0x7F != 9:
			raise ValueError("Packet has unexpected version number")
		
		packet = Packet()
		packet.encrypted = bool(encryption & 0x80)
		packet.destination_id = stream.u32()
		packet.source_id = stream.u32()
		packet.sequence_id = stream.u16()
		footer_size = stream.u8()
		packet.nonce = stream.u64()
		signature = stream.read(8)
		packet.payload = stream.read(stream.available() - footer_size)
		packet.footer = stream.read(footer_size)
		
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			packet.payload = aes.decrypt_and_verify(packet.payload, signature)
		
		return packet
	
	def encode(self, packet, addr):
		signature = bytes(16)
		payload = packet.payload
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			payload, signature = aes.encrypt_and_digest(packet.payload)
		
		stream = streams.StreamOut(self.settings)
		stream.u32(0x32AB9864)
		stream.u8((packet.encrypted << 7) | 9)
		stream.u32(packet.destination_id)
		stream.u32(packet.source_id)
		stream.u16(packet.sequence_id)
		stream.u8(len(packet.footer))
		stream.u64(packet.nonce)
		stream.write(signature)
		stream.write(payload)
		stream.write(packet.footer)
		return stream.get()


class PacketEncoderV11:
	def __init__(self, network, settings):
		self.network = network
		self.settings = settings
	
	def decode(self, data, addr):
		stream = streams.StreamIn(data, self.settings)
		if stream.u32() != 0x32AB9864:
			raise ValueError("Packet has wrong magic number")
		
		encryption = stream.u8()
		if encryption & 0x7F != 11:
			raise ValueError("Packet has unexpected version number")
		
		packet = Packet()
		packet.encrypted = bool(encryption & 0x80)
		packet.destination_id = stream.u16()
		packet.source_id = stream.u16()
		packet.sequence_id = stream.u16()
		footer_size = stream.u8()
		packet.nonce = stream.u64()
		signature = stream.read(8)
		packet.payload = stream.read(stream.available() - footer_size)
		packet.footer = stream.read(footer_size)
		
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			packet.payload = aes.decrypt_and_verify(packet.payload, signature)
		
		return packet
	
	def encode(self, packet, addr):
		signature = bytes(16)
		payload = packet.payload
		if packet.encrypted:
			nonce = self.network.generate_nonce(packet, addr)
			aes = AES.new(self.settings["crypto.session_key"], AES.MODE_GCM, nonce=nonce)
			payload, signature = aes.encrypt_and_digest(packet.payload)
		
		stream = streams.StreamOut(self.settings)
		stream.u32(0x32AB9864)
		stream.u8((packet.encrypted << 7) | 11)
		stream.u16(packet.destination_id & 0xFFFF)
		stream.u16(packet.source_id & 0xFFFF)
		stream.u16(packet.sequence_id)
		stream.u8(len(packet.footer))
		stream.u64(packet.nonce)
		stream.write(signature)
		stream.write(payload)
		stream.write(packet.footer)
		return stream.get()


class PacketEncoder:
	def __init__(self, network, settings):
		self.encoder = self.make_encoder(network, settings)
	
	def make_encoder(self, network, settings):
		version = settings["pia.version"]
		if version <= 506: return PacketEncoderV1(network, settings)
		elif 507 <= version <= 510: return PacketEncoderV2(network, settings)
		elif 511 <= version <= 517: return PacketEncoderV3(network, settings)
		elif 518 <= version <= 521: return PacketEncoderV4(network, settings)
		elif 523 <= version <= 526: return PacketEncoderV5(network, settings)
		elif 527 <= version <= 539: return PacketEncoderV9(network, settings)
		elif version == 616: return PacketEncoderV11(network, settings)
		else:
			raise ValueError("Unsupported pia version")
	
	def decode(self, data, addr): return self.encoder.decode(data, addr)
	def encode(self, packet, addr): return self.encoder.encode(packet, addr)


class PacketTransport:
	def __init__(self, network, settings, socket):
		self.encoder = PacketEncoder(network, settings)
		self.socket = socket
	
	async def send(self, packet, address):
		data = self.encoder.encode(packet, address)
		await self.socket.send(packet, address)
	
	async def receive(self):
		while True:
			data, address = await self.socket.recv()
			try:
				packet = self.encoder.decode(data, address)
				return packet, address
			except Exception:
				logger.warning("Received malformed packet")


class PacketHandler:
	def __init__(self, network, stations, settings, socket):
		self.transport = PacketTransport(network, settings, socket)
		self.stations = stations
		self.settings = settings
		
		self.nonce = random.randint(0, 0xFFFFFFFFFFFFFFFF)
		
		self.timer_base = time.monotonic()
		self.rtt = {}
	
	def next_nonce(self):
		self.nonce = (self.nonce + 1) & 0xFFFFFFFFFFFFFFFF
		return self.nonce
	
	def check_packet(self, packet, address):
		station = self.stations.find_by_address(address)
		if packet.connection_id > 2:
			if station is None or packet.connection_id != station.connection_id_in:
				logger.warning("Received packet with wrong connection id")
				return False
		
		if packet.destination_id != 0:
			local = self.stations.local_station()
			if packet.destination_id & 0xFFFF != local.variable_id & 0xFFFF:
				logger.warning("Received packet with wrong destination variable id")
				return False
		elif packet.footer:
			if len(packet.footer) % 2 != 0:
				logger.warning("Received packet with invalid footer")
				return False
			ids = struct.unpack(">%iH" %(len(packet.footer) // 2, packet.footer)
			if local.variable_id & 0xFFFF not in ids:
				logger.warning("Local variable id not found in packet footer")
				return False
		
		if packet.sequence_id != 0:
			if station is None or not station.check_sequence_id(packet.sequence_id):
				logger.warning("Received packet with wrong sequence id")
				return False
		
		current = int((time.monotonic() - self.timer_base) * 1000)
		self.rtt[address] = (current, packet.session_timer)
	
	async def send(self, packet, address):
		station = self.stations.find_by_address(address)
		if station:
			packet.connection_id = station.connection_id_out
			packet.destination_id = station.variable_id
			packet.sequence_id = station.next_sequence_id()
		
		local = self.stations.local_station()
		packet.source_id = local.variable_id
		
		current = int((time.monotonic() - self.timer_base) * 1000)
		packet.session_timer = current
		if address in self.rtt:
			local, remote = self.rtt[address]
			packet.rtt_timer = remote + (current - local)
		
		packet.nonce = self.next_nonce()
		await self.transport.send(packet, address)
	
	async def receive(self):
		while True:
			packet, address = await self.transport.receive()
			if self.check_packet(packet, address):
				return packet, address




class Station:
	def __init__(self):
		self.address = None
		
		self.index = 0xFD
		self.connection_id_out = 0
		self.connection_id_in = 0
		self.variable_id = 0
		self.constant_id = 0
		
		self.sequence_id_out = 0
		self.sequence_id_in = 0
	
	def next_sequence_id(self):
		if self.index != 0xFD:
			self.sequence_id_out = max((self.sequence_id_out + 1) & 0xFFFF, 1)
			return self.sequence_id_out
		return 0
	
	def check_sequence_id(self, sequence_id):
		if sequence_id == 0: return True
		
		diff = (sequence_id - self.sequence_id_in) & 0xFFFF
		if diff != 0 and diff < 0x8000:
			self.sequence_id_in = sequence_id
			return True
		return False


class StationTable:
	def __init__(self):
		self.stations = []
		self.create(None)
	
	def create(self, address):
		station = Station()
		station.address = address
		self.stations.append(station)
		return station
	
	def local_station(self):
		return self.find_by_address(None)
	
	def find_by_address(self, address):
		for station in self.stations:
			if station.address == address:
				return station
	
	def find_by_index(self, index):
		for station in self.stations:
			if station.index == index:
				return station
